# coding: utf-8
from django.conf import settings
from celery_balancer import logger
from celery_balancer.celery_api import CeleryApi
from celery_balancer.decision_executor import DecisionExecutor
from celery_balancer.decision_maker import DecisionMaker
from celery_balancer.helpers import nested_commit_on_success, smart_sleep
from celery_balancer.info_getter import InfoGetter
from celery_balancer.metrics.linux.metrics_api import LinuxMetricsApi
from celery_balancer.models import Queue, Worker
from celery_balancer.mq_api.rabbit_api import RabbitMQApi


class Balancer(object):
    """Балансировщик количества обработчиков у воркеров celery"""

    def __init__(self, error_log_file='', info_log_file='', logs_path='',
                 log_level=u'info'):
        logger.init_logging(error_log_file, info_log_file, logs_path,
                            log_level)
        self.celery_api = None
        self.mq_api = None
        self.metrics_api = None
        self.__init_api()
        self.decision_executor = None

    def run(self, stop_event):
        self.refresh_balancer_tables()
        self.bind_workers_to_queues()
        info_getter = InfoGetter(
            self.celery_api, self.mq_api, self.metrics_api
        )
        decision_maker = DecisionMaker()
        decision_executor = DecisionExecutor(self.celery_api)
        self.decision_executor = decision_executor

        while True:
            if stop_event.is_set():
                raise KeyboardInterrupt
            balancer_interval = self.get_balancer_interval()
            logger.info(u'Start new balancer loop.')
            info = info_getter.run()
            decision = decision_maker.run(info)
            decision_executor.run(decision)
            logger.info(u'Finish balancer loop. Now sleep for %s sec.' %
                        balancer_interval)
            smart_sleep(balancer_interval, stop_event)

    def stop_all_decisions_executions(self):
        """Остановка исполнений всех решений."""
        logger.info(u'stop_all_decisions_executions')
        self.decision_executor.stop_all_shrink_threads()

    def bind_workers_to_queues(self):
        queues_workers_dict = Queue.objects.get_queues_workers_names()
        for queue_name, queue_workers_names in queues_workers_dict.items():
            self.celery_api.bind_wokrkers_to_queue(
                queue_workers_names, queue_name
            )
        logger.info(u'bind_workers_to_queues')

    @staticmethod
    def get_balancer_interval():
        """Получение периода балансировки"""
        return Queue.objects.get_min_decision_interval()

    @nested_commit_on_success
    def refresh_balancer_tables(self):
        """
        Обновление таблиц балансировщика.
        Источник данных - settings.CELERY_QUEUES.
        """
        queues_names_to_delete = set()
        workers_names_to_delete = set()

        # Удалим из таблицы Queue записи, которых нет в CELERY_QUEUES
        db_queues_names = set(Queue.objects.get_names())
        settings_queues_names = set(settings.CELERY_QUEUES.keys())
        queues_names_to_delete.update(db_queues_names - settings_queues_names)

        for queue_name, queue_attrs in settings.CELERY_QUEUES.items():
            if not queue_attrs.get(u'workers_names', None):
                # Удалим из табилцы Queue записи,
                # для которых не заданы воркеры.
                queues_names_to_delete.add(queue_name)
            else:
                if Queue.objects.exists_by_name(queue_name):
                    # Удалим из таблицы Worker записи,
                    # которых нет для данной очереди в CELERY_QUEUES
                    queue = Queue.objects.get_by_name(queue_name)
                    db_queue_workers_names = set(queue.get_workers_names())
                    settings_workers_names = set(queue_attrs['workers_names'])
                    workers_names_to_delete.update(
                        db_queue_workers_names - settings_workers_names
                    )

                    # Создадим те записи, которые есть в CELERY_QUEUES для
                    # данной очереди, но нет в таблице Worker
                    workers_names_to_create = (
                        settings_workers_names - db_queue_workers_names
                    )
                    Worker.objects.create_workers_for_queue(
                        queue, workers_names_to_create
                    )
                else:
                    queue = Queue.objects.create(name=queue_name)
                    Worker.objects.create_workers_for_queue(
                        queue, queue_attrs['workers_names']
                    )

        Queue.objects.delete_by_names(queues_names_to_delete)
        Worker.objects.delete_by_names(workers_names_to_delete)
        logger.info(u'refresh_balancer_tables')

    def __init_api(self):
        """
        Задание используемых api.
        Поменяем здесь - применится для всего балансировщика.
        """
        self.celery_api = CeleryApi()
        self.mq_api = RabbitMQApi(self.celery_api)
        self.metrics_api = LinuxMetricsApi()