# coding: utf-8
import threading
import datetime
from celery_balancer import logger
from celery_balancer.helpers import smart_sleep
from celery_balancer.models import Worker


class DecisionExecutor(object):

    def __init__(self, celery_api):
        # self.shrink_threads = {
        #     worker_name:
        #         {u'thread': threading.Thread object,
        #          u'stop_event': threading.Event object}
        # }
        # self.grow_threads = {
        #     worker_name:
        #         {u'thread': threading.Thread object}
        # }
        self.shrink_threads = dict()
        self.grow_threads = dict()
        self.celery_api = celery_api

    def run(self, decision):
        """
        Исполнение решения.
        :param decision:
            decision = {
                'worker_name_1': {
                    'cur_proc_cnt': number,
                    'new_proc_cnt': number,
                    'force': bool, # может отсутствовать.
                },
                ...,
            }
        """
        if decision:
            logger.info(u'Start to execute decision in threads.')
        for worker_name, worker_decision in decision.items():
            worker_cur_proc_cnt = worker_decision[u'cur_proc_cnt']
            worker_new_proc_cnt = worker_decision[u'new_proc_cnt']

            if worker_new_proc_cnt > worker_cur_proc_cnt:
                if not worker_decision.get(u'force', False):
                    # Изменяем кол-во обработчиков только на 1, если не force.
                    worker_new_proc_cnt = worker_cur_proc_cnt + 1
                on = worker_cur_proc_cnt == 0
                grow_delta = worker_new_proc_cnt - worker_cur_proc_cnt
                self.__grow_proc_cnt(worker_name, grow_delta, on)
            elif worker_new_proc_cnt < worker_cur_proc_cnt:
                if not worker_decision.get(u'force', False):
                    # Изменяем кол-во обработчиков только на 1, если не force.
                    worker_new_proc_cnt = worker_cur_proc_cnt - 1
                off = worker_new_proc_cnt == 0
                shrink_delta = worker_cur_proc_cnt - worker_new_proc_cnt
                self.__shrink_proc_cnt(worker_name, shrink_delta,
                                       worker_cur_proc_cnt, off)

    def __grow_proc_cnt(self, worker_name, grow_delta, on):
        """
        Увеличение количества обработчиков на grow_delta
        :param worker_name: имя воркера
        :param on: происходит ли включение воркера (изменение кол-ва
                   обработчиков с нуля на большее число).
        """
        assert grow_delta > 0
        logger.info(u'Try to grow proc cnt for worker "%s"' % worker_name)
        if self.__alive_shrink_thread_for_worker_exists(worker_name):
            logger.info(u'Thread for shrink for worker "%s" already exixts. '
                        u'Stopping it.' % worker_name)
            self.__stop_shrink_thread(worker_name)
        if self.__alive_grow_thread_for_worker_exists(worker_name):
            logger.info(u'Worker "%s" already growing' % worker_name)
            return

        kwargs = {
            u'worker_name': worker_name,
            u'grow_delta': grow_delta,
            u'on': on,
        }
        grow_thread = threading.Thread(
            target=self.__grow_in_thread,
            kwargs=kwargs,
            name=u'grow_in_thread'
        )
        grow_thread.daemon = False
        grow_thread.start()

    def __grow_in_thread(self, worker_name, grow_delta, on):
        self.celery_api.grow_proc_cnt(grow_delta, worker_name)
        logger.info(u'Growed %s procs for worker "%s"' %
                    (grow_delta, worker_name))
        if on:
            Worker.objects.set_fields_values(
                {u'last_on_off_datetime': datetime.datetime.now()},
                (worker_name, )
            )

    def __shrink_proc_cnt(self, worker_name, shrink_delta,
                          worker_cur_proc_cnt, off):
        """Уменьшение кол-ва обработчиков на shrink_delta
        :param worker_name: имя воркера
        :param worker_cur_proc_cnt: текущее количество обработчиков воркера
                                    (до уменьшения)
        :param off: происходит ли выключение воркера (изменение кол-ва
                    обработчиков на ноль).
        """
        assert shrink_delta > 0
        logger.info(u'Try to shrink proc cnt for worker "%s"' % worker_name)
        if self.__alive_shrink_thread_for_worker_exists(worker_name):
            logger.info(u'Worker "%s" already shrinking' % worker_name)
            return

        queue_name = Worker.objects.get_by_name(worker_name).queue.name
        shrink_thread_stop_event = threading.Event()
        kwargs = {
            u'worker_name': worker_name,
            u'queue_name': queue_name,
            u'shrink_delta': shrink_delta,
            u'cur_proc_cnt': worker_cur_proc_cnt,
            u'off': off,
            u'stop_event': shrink_thread_stop_event,
        }
        shrink_thread = threading.Thread(
            target=self.__shrink_in_thread,
            kwargs=kwargs,
            name=u'shrink_in_thread'
        )
        shrink_thread.daemon = False
        self.shrink_threads[worker_name] = {
            u'thread': shrink_thread,
            u'stop_event': shrink_thread_stop_event,
        }
        shrink_thread.start()

    def __shrink_in_thread(self, worker_name, queue_name, shrink_delta,
                           cur_proc_cnt, off, stop_event):
        """
        :param worker_name: имя воркера
        :param queue_name: имя очереди
        :param shrink_delta: число, на которое нужно понизить
                             кол-во обработчиков
        :param cur_proc_cnt: текущее количество обработчиков воркера
                             (до уменьшения)
        :param off: происходит ли выключение воркера (изменение кол-ва
                    обработчиков на ноль).
        :param stop_event: объект threading.Event. Нужен для остановки.
        """
        total_proc_cnt = self.celery_api.get_total_proc_cnt().get(
            worker_name, 0
        )
        if total_proc_cnt == 0:
            logger.info(u'Already 0 procs. "%s"' % worker_name)
            return

        shrinked = self.__try_shrink(worker_name, shrink_delta, off)
        if not shrinked:
            logger.info(
                u'Worker "%s" is busy, starting shrink loop in thread.' %
                worker_name
            )
            self.__shrink_busy_worker(worker_name, queue_name, shrink_delta,
                                      cur_proc_cnt, off, stop_event)

    def __try_shrink(self, worker_name, n, off):
        """
        Попытка уменьшить количество обработчиков воркера на n.
        :return: 'ok' / ''
        """
        shrinked = False
        active_proc_cnt = self.celery_api.get_active_proc_cnt().get(
            worker_name, 0
        )
        total_proc_cnt = self.celery_api.get_total_proc_cnt().get(
            worker_name, 0
        )
        # Пока все воркеры заняты работой,
        # нет смысла пытаться уменьшить их кол-во.
        if total_proc_cnt - active_proc_cnt >= n:
            shrinked = self.celery_api.shrink_proc_cnt(
                n, worker_name, reply=True
            )

        if shrinked:
            logger.info(u'Shrinked %s procs for worker "%s"' %
                        (n, worker_name))
            if off:
                Worker.objects.set_fields_values(
                    {u'last_on_off_datetime': datetime.datetime.now()},
                    (worker_name, )
                )

        return shrinked

    def __shrink_busy_worker(self, worker_name, queue_name, shrink_delta,
                             cur_proc_cnt, off, stop_event):
        """
        Уменьшение кол-ва обработчиков воркера,
        у которого нет свободных обработчиков.
        Такой воркер надо сначала отписать от очереди, а потом ждать,
        пока осовободится нужное количество обработчиков. Далее уменьшить
        кол-во обработчиков и снова подписать воркер на очередь.

        :param worker_name: имя воркера
        :param queue_name: имя очереди
        :param shrink_delta: число, на которое нужно понизить
                             кол-во обработчиков
        :param cur_proc_cnt: текущее количество обработчиков воркера
                             (до уменьшения)
        :param off: происходит ли выключение воркера (изменение кол-ва
                    обработчиков на ноль).
        :param stop_event: объект threading.Event. Нужен для остановки.
        """
        shrinked = False
        Worker.objects.set_fields_values(
            {u'expected_proc_cnt': cur_proc_cnt - shrink_delta},
            (worker_name, )
        )

        self.celery_api.unbind_worker_from_queue(worker_name, queue_name)
        while not shrinked and not stop_event.is_set():
            shrinked = self.__try_shrink(worker_name, shrink_delta, off)
            if not shrinked:
                smart_sleep(10, stop_event)
        self.celery_api.bind_wokrkers_to_queue([worker_name], queue_name)

        Worker.objects.set_fields_values(
            {u'expected_proc_cnt': None},
            (worker_name, )
        )
        if not shrinked and stop_event.is_set():
            logger.info(u'Shrink thread stopped for worker "%s"'
                        % worker_name)

    def __alive_shrink_thread_for_worker_exists(self, worker_name):
        """Если ли живой тред для воркера на понижение числа обработчиков."""
        return self.__alive_thread_for_worker_exists(worker_name,
                                                     self.shrink_threads)

    def __alive_grow_thread_for_worker_exists(self, worker_name):
        """Если ли живой тред для воркера на повышение числа обработчиков."""
        return self.__alive_thread_for_worker_exists(worker_name,
                                                     self.grow_threads)

    @staticmethod
    def __alive_thread_for_worker_exists(worker_name, threads_dict):
        """Если ли живой тред для воркера."""
        thread_info = threads_dict.get(worker_name, None)
        thread = thread_info and thread_info[u'thread']
        return bool(thread and thread.is_alive())

    def __stop_shrink_thread(self, worker_name):
        thread_info = self.shrink_threads.get(worker_name, None)
        if thread_info:
            stop_event = thread_info[u'stop_event']
            stop_event.set()

    def stop_all_shrink_threads(self):
        for worker_name in self.shrink_threads:
            self.__stop_shrink_thread(worker_name)
